use log::{trace, warn};
use std::{
    os::unix::process::ExitStatusExt,
    process::{Child, Command}, path::{Path, PathBuf},
};

use crate::pidfd::{CreatePidFd, PidFd};

pub struct Subservice {
    path: PathBuf,
    child: Child,
    pidfd: PidFd,
}

impl Subservice {
    pub fn start<P: AsRef<Path>, S: AsRef<str>>(helper_dir: P, binary_name: S) -> anyhow::Result<Self> {
        let path = helper_dir.as_ref().join(binary_name.as_ref());
        Self::start_internal(path)
    }

    fn start_internal<P: AsRef<Path>>(path: P) -> anyhow::Result<Self> {
        let child = Command::new(path.as_ref())
            .spawn()?;
        let pidfd = child.create_pidfd()?;

        Ok(Self {
            path: path.as_ref().to_path_buf(),
            child,
            pidfd,
        })
    }

    pub fn pidfd(&self) -> &PidFd {
        &self.pidfd
    }

    pub fn handle_quit(mut self) -> anyhow::Result<Option<Self>> {
        if let Some(status) = self.child.try_wait().ok().flatten() {
            if !status.success() {
                warn!("{} service exited abnormally ({}{}); restarting", self.path.to_string_lossy(),
                    status.code().map(|c| format!("code {}", c)).unwrap_or("".to_string()),
                    status.signal().map(|s| format!("signal {}", s)).unwrap_or("".to_string())
                );
                Self::start_internal(&self.path).map(Some)
            } else {
                trace!("{} service exited normally", self.path.to_string_lossy());
                Ok(None)
            }
        } else {
            trace!("{} service didn't seem to actually quit", self.path.to_string_lossy());
            Ok(Some(self))
        }
    }
}

impl Drop for Subservice {
    fn drop(&mut self) {
        let _ = self.child.kill();
        let _ = self.child.try_wait();
    }
}

pub struct Subservices {
    dbus_service: Option<Subservice>,
    systemd_service: Option<Subservice>,
}

impl Subservices {
    pub fn start_all<P: AsRef<Path>>(helper_dir: P) -> anyhow::Result<Subservices> {
        let dbus_service = Subservice::start(helper_dir.as_ref(), "bscreensaver-dbus-service")?;
        let systemd_service = Subservice::start(helper_dir, "bscreensaver-systemd")?;
        Ok(Self {
            dbus_service: Some(dbus_service),
            systemd_service: Some(systemd_service),
        })
    }

    pub fn stop_all(&mut self) {
        let _ = self.dbus_service.take();
        let _ = self.systemd_service.take();
    }

    pub fn dbus_service(&self) -> Option<&Subservice> {
        self.dbus_service.as_ref()
    }

    pub fn systemd_service(&self) -> Option<&Subservice> {
        self.systemd_service.as_ref()
    }

    pub fn handle_quit(&mut self) -> anyhow::Result<()> {
        if let Some(dbus_service) = self.dbus_service.take() {
            self.dbus_service = dbus_service.handle_quit()?;
        }
        if let Some(systemd_service) = self.systemd_service.take() {
            self.systemd_service = systemd_service.handle_quit()?;
        }
        Ok(())
    }
}
