use std::{env, fmt, process::Command};

const GDM_CMD: &[&str] = &[ "gdmflexiserver", "-ls" ];
const KDM_CMD: &[&str] = &[ "kdmctl", "reserve" ];
const LIGHTDM_CMD: &[&str] = &[ "dm-tool", "switch-to-greeter" ];
const LXDM_CMD: &[&str] = &[ "lxdm", "-c", "USER_SWITCH" ];

#[derive(Debug, Clone, PartialEq)]
pub enum NewLoginCommand {
    Auto,
    DisplayManagerDBus,
    Gdm,
    Kdm,
    LightDm,
    Lxdm,
    Custom(String),
    Disabled,
}

impl NewLoginCommand {
    pub fn run(&self) -> anyhow::Result<()> {
        match self {
            Self::Auto => auto_run(),
            Self::DisplayManagerDBus => call_dbus(),
            Self::Gdm => run(GDM_CMD),
            Self::Kdm => run(KDM_CMD),
            Self::LightDm => run(LIGHTDM_CMD),
            Self::Lxdm => run(LXDM_CMD),
            Self::Custom(cmd) => shell_words::split(cmd.as_str())
                .map_err(|err| err.into())
                .and_then(|argvec| {
                    let argv: Vec<&str> = argvec.iter().map(|s| s.as_str()).collect();
                    run(&argv).map_err(|err| err.into())
                }),
            Self::Disabled => Ok(()),
        }
    }

    pub fn as_str(&self) -> &'static str {
        match self {
            Self::Auto => "Auto",
            Self::DisplayManagerDBus => "DBus (desktop-agnostic)",
            Self::Gdm => "GDM (GNOME)",
            Self::Kdm => "KDM (KDE)",
            Self::LightDm => "LightDM",
            Self::Lxdm => "LXDM (LXDE)",
            Self::Custom(_) => "Custom",
            Self::Disabled => "Disabled",
        }
    }

    pub fn ord(&self) -> u32 {
        match self {
            Self::Auto => 0,
            Self::DisplayManagerDBus => 1,
            Self::Gdm => 2,
            Self::Kdm => 3,
            Self::LightDm => 4,
            Self::Lxdm => 5,
            Self::Custom(_) => 6,
            Self::Disabled => 7,
        }
    }
}

impl fmt::Display for NewLoginCommand {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.as_str())
    }
}

impl TryFrom<u32> for NewLoginCommand {
    type Error = anyhow::Error;
    fn try_from(value: u32) -> Result<Self, Self::Error> {
        match value {
            0 => Ok(Self::Auto),
            1 => Ok(Self::DisplayManagerDBus),
            2 => Ok(Self::Gdm),
            3 => Ok(Self::Kdm),
            4 => Ok(Self::LightDm),
            5 => Ok(Self::Lxdm),
            6 => Ok(Self::Custom("".to_string())),
            7 => Ok(Self::Disabled),
            other => Err(anyhow::anyhow!("Unknown ordinal for NewLoginCommand: {}", other)),
        }
    }
}

impl TryFrom<&toml::Value> for NewLoginCommand {
    type Error = anyhow::Error;
    fn try_from(value: &toml::Value) -> Result<Self, Self::Error> {
        if let Some(s) = value.as_str() {
            return match s {
                "auto" => Ok(Self::Auto),
                "dbus" => Ok(Self::DisplayManagerDBus),
                "gdm" => Ok(Self::Gdm),
                "kdm" => Ok(Self::Kdm),
                "lightdm" => Ok(Self::LightDm),
                "lxdm" => Ok(Self::Lxdm),
                "disabled" => Ok(Self::Disabled),
                other => Err(anyhow::anyhow!("Unknown value '{}' for new login command", other)),
            };
        }

        if let Some(arr) = value.as_array() {
            let argv = arr.into_iter()
                .map(|v| v.as_str().map(|s| s.to_string()).ok_or_else(|| anyhow::anyhow!("Custom command must be made of strings")))
                .collect::<anyhow::Result<Vec<String>>>()?;
            return Ok(Self::Custom(shell_words::join(argv)))
        }

        Err(anyhow::anyhow!("Unknown value for new login command"))
    }
}

impl TryInto<toml::Value> for NewLoginCommand {
    type Error = anyhow::Error;
    fn try_into(self) -> Result<toml::Value, Self::Error> {
        match self {
            Self::Auto => Ok(toml::Value::String("auto".to_string())),
            Self::DisplayManagerDBus => Ok(toml::Value::String("dbus".to_string())),
            Self::Gdm => Ok(toml::Value::String("gdm".to_string())),
            Self::Kdm => Ok(toml::Value::String("kdm".to_string())),
            Self::LightDm => Ok(toml::Value::String("lightdm".to_string())),
            Self::Lxdm => Ok(toml::Value::String("lxdm".to_string())),
            Self::Custom(cmd) => shell_words::split(cmd.as_str())
                .map_err(|err| err.into())
                .map(|words| toml::Value::Array(words.into_iter().map(|s| toml::Value::String(s)).collect())),
            Self::Disabled => Ok(toml::Value::String("disabled".to_string())),
        }
    }
}

fn auto_run() -> anyhow::Result<()> {
    if let Ok(_) = call_dbus() {
        return Ok(())
    }

    if let Some(session_desktop) = env::var("XDG_SESSION_DESKTOP").map(|s| s.to_lowercase()).ok() {
        match session_desktop.as_str() {
            "gnome" | "gdm" => run(GDM_CMD),
            "kde" | "kdm" => run(KDM_CMD),
            "lightdm-session" => run(LIGHTDM_CMD),
            "lxde" | "lxdm" => run(LXDM_CMD),
            other => Err(anyhow::anyhow!("Couldn't determine how to start a new login session for session type {}", other)),
        }
    } else {
        Err(anyhow::anyhow!("Couldn't determine how to start a new login session"))
    }
}

fn call_dbus() -> anyhow::Result<()> {
    let seat_path = env::var("XDG_SEAT_PATH")
        .unwrap_or_else(|_| "/org/freedesktop/DisplayManager/Seat0".to_string());
    let bus = zbus::blocking::Connection::system()?;
    let proxy = zbus::blocking::Proxy::new(
        &bus,
        "org.freedesktop.DisplayManager",
        seat_path,
        "org.freedesktop.DisplayManager.Seat"
    )?;
    proxy.call("SwitchToGreeter", &())?;
    Ok(())
}

fn run(argv: &[&str]) -> anyhow::Result<()> {
    Command::new(argv[0])
        .args(if argv.len() > 1 { &argv[1..argv.len()-1] } else { &[] })
        .spawn()?;
    Ok(())
}
